package meerkat.bulletinboard;

import com.google.common.util.concurrent.*;
import com.google.protobuf.ByteString;
import meerkat.bulletinboard.callbacks.GetRedundancyFutureCallback;
import meerkat.bulletinboard.callbacks.PostMessageFutureCallback;
import meerkat.bulletinboard.callbacks.ReadMessagesFutureCallback;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Digest;
import meerkat.crypto.concrete.SHA256Digest;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting;

import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Created by Arbel Deutsch Peled on 05-Dec-15.
 * Thread-based implementation of a Bulletin Board Client.
 * Features:
 * 1. Handles tasks concurrently.
 * 2. Retries submitting
 */
public class ThreadedBulletinBoardClient implements BulletinBoardClient {

    private final static int THREAD_NUM = 10;
    ListeningExecutorService listeningExecutor;

    private Digest digest;

    private List<String> meerkatDBs;
    private String postSubAddress;
    private String readSubAddress;

    private final static int READ_MESSAGES_RETRY_NUM = 1;

    private int minAbsoluteRedundancy;

    /**
     * Stores database locations and initializes the web Client
     * Stores the required minimum redundancy.
     * Starts the Thread Pool.
     * @param clientParams contains the required information
     */
    @Override
    public void init(Voting.BulletinBoardClientParams clientParams) {

        meerkatDBs = clientParams.getBulletinBoardAddressList();

        minAbsoluteRedundancy = (int) (clientParams.getMinRedundancy() * meerkatDBs.size());

        listeningExecutor =  MoreExecutors.listeningDecorator(Executors.newFixedThreadPool(THREAD_NUM));

        digest = new SHA256Digest();

    }

    /**
     * Post message to all DBs
     * Retry failed DBs
     * @param msg is the message,
     * @return the message ID for later retrieval
     * @throws CommunicationException
     */
    @Override
    public MessageID postMessage(BulletinBoardMessage msg, ClientCallback<?> callback){

        // Create job
        BulletinClientJob job = new BulletinClientJob(meerkatDBs, minAbsoluteRedundancy, BulletinClientJob.JobType.POST_MESSAGE, msg, -1);

        // Submit job and create callback
        Futures.addCallback(listeningExecutor.submit(new BulletinClientWorker(job)), new PostMessageFutureCallback(listeningExecutor, callback));

        // Calculate the correct message ID and return it
        digest.reset();
        digest.update(msg.getMsg());
        return MessageID.newBuilder().setID(ByteString.copyFrom(digest.digest())).build();
    }

    /**
     * Access each database and search for a given message ID
     * Return the number of databases in which the message was found
     * Only try once per DB
     * Ignore communication exceptions in specific databases
     * @param id is the requested message ID
     * @return the number of DBs in which retrieval was successful
     */
    @Override
    public void getRedundancy(MessageID id, ClientCallback<Float> callback) {

        // Create job
        BulletinClientJob job = new BulletinClientJob(meerkatDBs, minAbsoluteRedundancy, BulletinClientJob.JobType.GET_REDUNDANCY, id, 1);

        // Submit job and create callback
        Futures.addCallback(listeningExecutor.submit(new BulletinClientWorker(job)), new GetRedundancyFutureCallback(listeningExecutor, callback));

    }

    /**
     * Go through the DBs and try to retrieve messages according to the specified filter
     * If at the operation is successful for some DB: return the results and stop iterating
     * If no operation is successful: return null (NOT blank list)
     * @param filterList return only messages that match the filters (null means no filtering).
     * @return
     */
    @Override
    public void readMessages(MessageFilterList filterList, ClientCallback<List<BulletinBoardMessage>> callback) {

        // Create job
        BulletinClientJob job = new BulletinClientJob(meerkatDBs, minAbsoluteRedundancy, BulletinClientJob.JobType.READ_MESSAGES,
                filterList, READ_MESSAGES_RETRY_NUM);

        // Submit job and create callback
        Futures.addCallback(listeningExecutor.submit(new BulletinClientWorker(job)), new ReadMessagesFutureCallback(listeningExecutor, callback));

    }

    @Override
    public void close() {
        try {
            listeningExecutor.shutdown();
            while (! listeningExecutor.isShutdown()) {
                listeningExecutor.awaitTermination(10, TimeUnit.SECONDS);
            }
        } catch (InterruptedException e) {
            System.err.println(e.getCause() + " " + e.getMessage());
        }
    }

}
