package JointFeldmanProtocol;

import Communication.User;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import ShamirSecretSharing.Polynomial;
import com.google.protobuf.ByteString;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 3/14/2016.
 */
// TODO: Lots of comments...
// TODO: User Channel instead of User
public class DistributedKeyGeneration extends VerifiableSecretSharing {
    public enum ComplaintState {
        /**
         * No complaints, no response required at this point.
         */
        OK,

        /**
         * Party received complaint, waiting for response from party
         */
        Waiting,

        /**
         * Party gave invalid answer to conplaint.
         */
        Disqualified,

        /**
         * Party received complaint, gave valid answer.
         */
        NonDisqualified
    }

    /**
     * My share id.
     */
    protected final int id;

    /**
     * All parties participating in key generation.
     * parties[id-1] has my info.
     */
    private DistributedKeyGenerationParty[] parties;


    // TODO: Copy comment
    public DistributedKeyGeneration(int t, int n, BigInteger zi, Random random, BigInteger q, BigInteger g
            , Group<BigInteger> group, int id)  {
        super(t, n, zi, random, q, g,group);
        this.id = id;
        this.parties = new DistributedKeyGenerationParty[n];
        for (int i = 1; i <= n ; i++){
            this.parties[i - 1] = new DistributedKeyGenerationParty(i,n,t);
        }
        this.parties[id - 1].share = getShare(id);
    }

    protected void setParties(DistributedKeyGenerationParty[] parties){
        this.parties = parties;
    }

    protected DistributedKeyGenerationParty[] getParties(){
        return parties;
    }

    /**
     * stage1.1 according to the protocol
     *  Pi broadcasts Aik for k = 0,...,t.
     */
    public void broadcastCommitments(User user){
        broadcastCommitments(user,commitmentsArray);
    }

    public void broadcastCommitments(User user, BigInteger[] commitments){
        DKGMessages.CommitmentMessage commitmentMessage;
        for (int k = 0; k <= t ; k++){
            commitmentMessage = DKGMessages.CommitmentMessage.newBuilder()
                    .setCommitment(ByteString.copyFrom(commitments[k].toByteArray()))
                    .setK(k)
                    .build();
            user.broadcast(DKGMessages.Mail.Type.COMMITMENT, commitmentMessage);
        }
    }

    /**
     * Send user j her secret share (of my polynomial)
     * @param user
     * @param j
     */
    public void sendSecret(User user, int j){
        ByteString secret = ByteString.copyFrom(getShare(j).y.toByteArray());
        user.send(j, DKGMessages.Mail.Type.SECRET,
                DKGMessages.SecretMessage.newBuilder()
                        .setI(id)
                        .setJ(j)
                        .setSecret(secret)
                        .build());
    }

    /**
     * stage1.2 according to the protocol
     *  Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    public void sendSecrets(User user){
        for (int j = 1; j <= n ; j++){
            if(j != id){
                sendSecret(user,j);
            }
        }
    }

    /**
     * TODO: comment
     * @param i
     * @return
     */
    public boolean isValidSecret(int i){
        DistributedKeyGenerationParty party = parties[i - 1];
        return isValidSecret(party.share,party.commitments,id);
    }

    /**
     * TODO: Move to VerifiableSecretSharing
     * @param secret
     * @param commitments
     * @param j
     * @return computeVerificationValue(j,commitments,group) == g ^ secret.y mod q
     */
    public boolean isValidSecret(Polynomial.Point secret, BigInteger[] commitments, int j){
        try{
            BigInteger v = computeVerificationValue(j,commitments,group);
            return group.multiply(g,secret.y).equals(v);
        }
        catch (NullPointerException e){
            return false;
        }
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received (using isValidShare)
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     */
    public void broadcastComplaints(User user){
        for (int i = 1; i <= n ; i++ ){
            if(i != id && !isValidSecret(i)) {
                broadcastComplaint(user,i);
            }
        }
    }

    private void broadcastComplaint(User user, int i){
        //message = new Message(Type.Complaint, j)
        DKGMessages.IDMessage complaint = DKGMessages.IDMessage.newBuilder()
                .setId(i)
                .build();
        user.broadcast(DKGMessages.Mail.Type.COMPLAINT, complaint);
    }

    public void broadcastComplaintAnswer(User user, int j){
        user.broadcast(DKGMessages.Mail.Type.ANSWER, DKGMessages.SecretMessage.newBuilder()
                .setI(id)
                .setJ(j)
                .setSecret(ByteString.copyFrom(getShare(j).y.toByteArray()))
                .build());
    }

    /**
     * stage3.1 according to the protocol
     *   if more than t players complain against a player Pi he is disqualified.
     */
    public void answerAllComplainingPlayers(User user){
        ComplaintState[] complaints = parties[id - 1].complaints;
        for (int i = 1; i <= n ; i++) {
            switch (complaints[i - 1]) {
                case Waiting:
                    broadcastComplaintAnswer(user,i);
                    break;
                default:
                    break;
            }
        }
    }

    /**
     * stage3.2 according to the protocol
     *  if any of the revealed shares fails the verification test, player Pi is disqualified.
     *  set QUAL to be the set of non-disqualified players.
     */
    public Set<Integer> calcQUAL(){
        Set<Integer> QUAL = new HashSet<Integer>();
        boolean nonDisqualified;
        int counter;
        for (int i = 1; i <= n; i++){
            ComplaintState[] complaints = parties[i - 1].complaints;
            nonDisqualified = true;
            counter = 0;
            for (int j = 1; j <= n; j++){
                switch (complaints[j - 1]) {
                    case OK:
                        break;
                    case NonDisqualified:
                        // TODO: Add test for false complaint
                        counter++;
                        break;
                    default:
                        nonDisqualified = false;
                        break;
                }
                if(!nonDisqualified)
                    break;
            }
            if(nonDisqualified && counter <= t){
                QUAL.add(i);
            }
        }
        return QUAL;
    }

    /**
     * compute Y, the commitment to the final public key (includes only qualifying set)
     * stage4.1 according to the protocol
     *   public value y is computed as y = multiplication of yi mod p for i in QUAL
     */
    public BigInteger calcY(Set<Integer> QUAL){
        BigInteger y = group.zero();
        for (int i : QUAL) {
            y = group.add(y , parties[i - 1].commitments[0]);
        }
        return y;
    }

    /**
     * TODO: better comment.
     * stage4.2 according to the protocol
     *  public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     */
    public BigInteger[] calcCommitments(Set<Integer> QUAL){
        BigInteger[] commitments = new BigInteger[t + 1];
        Arrays.fill(commitments,group.zero());
        for (int i : QUAL) {
            for (int k = 0; k <= t; k++){
                commitments[k] = group.add(commitments[k], parties[i - 1].commitments[k]);
            }
        }
        return commitments;
    }

    /**
     * TODO: better comment.
     * stage4.3 according to the protocol
     *  Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    public Polynomial.Point calcShare(Set<Integer> QUAL){
        BigInteger xj = BigInteger.ZERO;
        for (int i : QUAL) {
            xj = xj.add(parties[i - 1].share.y);
        }
        return new Polynomial.Point(BigInteger.valueOf(id) , xj.mod(q));
    }

    /**
     * getter
     * @return id
     */
    public int getId() {
        return id;
    }

}
