package SecureDistributedKeyGenerationForDiscreteLogBasedCryptosystem;

import Arithmetics.Arithmetic;
import Arithmetics.Fp;
import Communication.Network;
import JointFeldmanProtocol.DistributedKeyGeneration;
import JointFeldmanProtocol.DistributedKeyGenerationUserImpl;
import ShamirSecretSharing.Polynomial;
import ShamirSecretSharing.SecretSharing;
import com.google.protobuf.Message;
import meerkat.protobuf.DKGMessages;

import java.math.BigInteger;

/**
 * Created by Tzlil on 3/16/2016.
 */
public class SecureDistributedKeyGenerationUserImpl extends DistributedKeyGenerationUserImpl {

    protected SecureDistributedKeyGenerationParty[] parties;
    protected final SecureDistributedKeyGeneration sdkg;
    private Arithmetic<BigInteger> arithmetic;
    private boolean isStage4;

    public SecureDistributedKeyGenerationUserImpl(SecureDistributedKeyGeneration sdkg, Network network) {
        super(sdkg, network,new SecureDistributedKeyGenerationMailHandler(null));
        this.sdkg = sdkg;
        this.messageHandler = new MessageHandler();
        this.user.setMessageHandler(this.messageHandler);
        this.parties = sdkg.getParties();
        this.arithmetic = new Fp(sdkg.getQ());
        this.isStage4 = false;
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Cik=Aik*Bik for k = 0,...,t.
     *  2. Pi computes the shares Sij,Sij' for j = 1,...,n and sends Sij,Sij' secretly to Pj.
     */
    @Override
    protected void stage1() {
        sdkg.broadcastVerificationValues(user);
        sdkg.sendSecrets(user);
    }

    @Override
    protected void waitUntilStageOneCompleted(){
        super.waitUntilStageOneCompleted();
        // save the received commitments as verification values
        BigInteger[] temp;
        for (int i = 0 ; i < n; i++){
            temp = parties[i].verifiableValues;
            parties[i].verifiableValues = parties[i].commitments;
            parties[i].commitments = temp;
        }
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares,sharesT he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    @Override
    protected void stage2(){
        sdkg.broadcastComplaints(user);
        //broadcast done message after all complaints
        DKGMessages.EmptyMessage doneMessage = DKGMessages.EmptyMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);
    }

    // TODO: ??
    private void resolveQualifyingPublicKey(){
        sdkg.broadcastCommitments(user);
        // wait until all parties in QUAL broadcast their commitments or aborted
        // TODO: in main run loop
        for (int i:QUAL) {
            for(int k = 0; k <= t; k++) {
                while (parties[i - 1].commitments[k] == null && !parties[i - 1].aborted) {
                    try {
                        Thread.sleep(SleepTime);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
        sdkg.computeAndBroadcastComplaints(user,QUAL);
        //broadcast done message after all complaints
        DKGMessages.EmptyMessage doneMessage = DKGMessages.EmptyMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);

        // wait until all parties in QUAL done or aborted
        for (int i:QUAL) {
            while (!parties[i - 1].ysDoneFlag && !parties[i - 1].aborted) {
                try {
                    Thread.sleep(SleepTime);
                } catch (InterruptedException e) {
                    // do nothing
                }
            }
        }

        // broadcast i private secret foreach i in QUAL that aborted
        for (int i:QUAL) {
            if(parties[i - 1].aborted){
                sdkg.broadcastAnswer(user, parties[i - 1].share, parties[i - 1].shareT, i);
            }
        }
        // wait until at least t + 1 secrets will received foreach i in QUAL that aborted
        for (int i:QUAL) {
            if(parties[i - 1].aborted){
                while (parties[i - 1].restoreSharesSet.size() <= t) {
                    try {
                        Thread.sleep(SleepTime);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }

        // restore necessary information
        for (int i = 0; i < n ; i++) {
            if(parties[i].restoreSharesSet.isEmpty()){
                continue;
            }
            Polynomial.Point[] shares = new Polynomial.Point[t + 1];
            int j = 0;
            for (Polynomial.Point share: parties[i].restoreSharesSet){
                shares[j++] = share;
                if (j >= shares.length){
                    break;
                }
            }
            Polynomial polynomial = SecretSharing.recoverPolynomial(shares,arithmetic);
            BigInteger[] coefficients = polynomial.getCoefficients();
            for (int k = 0 ; k <= t; k++){
                parties[i].commitments[k] = group.multiply(g,coefficients[k]);
            }
            parties[i].share = new Polynomial.Point(BigInteger.valueOf(id),polynomial);
        }
    }

    /**
     * notifies mail handler that stage 4 was started
     */
    protected void setStage4(){
        this.isStage4 = true;
        SecureDistributedKeyGenerationMailHandler handler =
                (SecureDistributedKeyGenerationMailHandler)user.getMailHandler();
        handler.setStage4(true);
    }

    @Override
    protected void stage4() {
        setStage4();
        resolveQualifyingPublicKey();
        super.stage4();
    }

    private class MessageHandler extends DistributedKeyGenerationUserImpl.MessageHandler{

        /**
         * as in super, with extension to double secret message
         */
        protected boolean isValidSecretMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {
            DKGMessages.SecretMessage secretMessage = DKGMessages.SecretMessage.newBuilder()
                    .setI(doubleSecretMessage.getI())
                    .setJ(doubleSecretMessage.getJ())
                    .setSecret(doubleSecretMessage.getSecret())
                    .build();
            return super.isValidSecretMessage(sender,isBroadcast,secretMessage);
        }

        /**
         * as in super, with extension to double secret message
         */
        @Override
        public void handleSecretMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.DoubleSecretMessage doubleSecretMessage = (DKGMessages.DoubleSecretMessage)message;
            if (isValidSecretMessage(sender,isBroadcast,doubleSecretMessage)) {
                int i = doubleSecretMessage.getI();
                parties[i - 1].share = extractSecret(id, doubleSecretMessage.getSecret());
                parties[i - 1].shareT = extractSecret(id, doubleSecretMessage.getSecretT());
            }
        }

        /**
         * if !isStage4  as super, with extension to double secret message
         * else answer message is valid if:
         *  1. it was received in broadcast chanel
         *  2. secret.j == sender
         *  3. QUAL contains i and j
         */
        protected boolean isValidAnswerMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {
            if(!isStage4) {
                DKGMessages.SecretMessage secretMessage = DKGMessages.SecretMessage.newBuilder()
                        .setI(doubleSecretMessage.getI())
                        .setJ(doubleSecretMessage.getJ())
                        .setSecret(doubleSecretMessage.getSecret())
                        .build();
                return super.isValidAnswerMessage(sender, isBroadcast, secretMessage);
            }else{
                int i = doubleSecretMessage.getI();
                int j = doubleSecretMessage.getJ();
                return isBroadcast && j == sender && parties[i -1].aborted && !parties[j - 1].aborted
                        && QUAL.contains(i) && QUAL.contains(j);
            }
        }

        /**
         * if !isStage4 as super, with extension to double secret message
         * else saves secret
         */
        @Override
        public void handleAnswerMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.DoubleSecretMessage doubleSecretMessage = (DKGMessages.DoubleSecretMessage)message;
            if(isValidAnswerMessage(sender,isBroadcast,doubleSecretMessage)) {
                int i = doubleSecretMessage.getI();
                int j = doubleSecretMessage.getJ();
                Polynomial.Point secret = extractSecret(j, doubleSecretMessage.getSecret());
                Polynomial.Point secretT = extractSecret(j, doubleSecretMessage.getSecretT());
                if (!isStage4) {
                    if (sdkg.isValidShare(secret, secretT, parties[j - 1].verifiableValues, i)) {
                        parties[i - 1].complaints[j - 1] = DistributedKeyGeneration.ComplaintState.NonDisqualified;

                    } else {
                        parties[i - 1].complaints[j - 1] = DistributedKeyGeneration.ComplaintState.Disqualified;
                    }
                    if(j == id){
                        parties[i - 1].share = secret;
                        parties[i - 1].shareT = secretT;
                    }
                } else if (sdkg.isValidShare(secret, secretT, parties[j - 1].verifiableValues, i)) {
                    // TODO: Check that this is ok
                    parties[i - 1].restoreSharesSet.add(secret);
                }
            }
        }

        /**
         * as in super with respect to protocol stage
         */
        @Override
        protected boolean isValidDoneMessage(int sender, boolean isBroadcast) {
            if(!isStage4) {
                return super.isValidDoneMessage(sender, isBroadcast);
            }else{
                return isBroadcast &&  !parties[sender - 1].ysDoneFlag;
            }
        }

        /**
         * as in super with respect to protocol state
         */
        @Override
        public void handleDoneMessage(int sender, boolean isBroadcast, Message message) {
            if(!isStage4)
                super.handleDoneMessage(sender, isBroadcast, message);
            else{
                if(isValidDoneMessage(sender,isBroadcast)) {
                    parties[sender - 1].ysDoneFlag = true;
                }
            }
        }

        /**
         * use only in stage4
         * complaint message is valid if:
         *  1. it was received in broadcast chanel
         *  2. secret.j == sender
         *  3. QUAL contains i and j
         */
        protected boolean isValidComplaintMessage(int sender, boolean isBroadcast,
                                                  DKGMessages.DoubleSecretMessage complaintMessage){
            int i = complaintMessage.getI();
            int j = complaintMessage.getJ();
            return isBroadcast && j == sender && QUAL.contains(i) && QUAL.contains(j);
        }

        /**
         * if !isStage4 as in super
         * else if secret,secretT are valid with respect to verifiableValues but
         *      secret is not valid with respect to commitments then
         *      marks i as aborted
         */
        @Override
        public void handleComplaintMessage(int sender, boolean isBroadcast, Message message) {
            if(!isStage4) {
                super.handleComplaintMessage(sender, isBroadcast, message);
            }else {
                DKGMessages.DoubleSecretMessage ysComplaintMessage =(DKGMessages.DoubleSecretMessage)message;
                if (isValidComplaintMessage(sender,isBroadcast,ysComplaintMessage)) {
                    int i = ysComplaintMessage.getI();
                    int j = ysComplaintMessage.getJ();
                    Polynomial.Point secret = extractSecret(i,ysComplaintMessage.getSecret());
                    Polynomial.Point secretT = extractSecret(i,ysComplaintMessage.getSecretT());
                    if (sdkg.isValidShare(secret, secretT, parties[i - 1].verifiableValues, j)
                            && !dkg.isValidSecret(secret,parties[i - 1].commitments, j)) {
                        parties[i - 1].aborted = true;
                    }
                }
            }
        }
    }
}
