package JointFeldmanProtocol;

import Arithmetics.Arithmetic;
import Arithmetics.Fp;
import Communication.Network;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import ShamirSecretSharing.Polynomial;
import ShamirSecretSharing.SecretSharing;
import UserInterface.DistributedKeyGenerationUser;
import Utils.GenerateRandomPrime;
import org.factcenter.qilin.primitives.Group;
import org.factcenter.qilin.primitives.concrete.Zpstar;
import org.junit.Before;
import org.junit.Test;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 3/21/2016.
 */
public class DKGTest {

    int tests = 10;
    BigInteger p = GenerateRandomPrime.SafePrime100Bits;
    BigInteger q = p.subtract(BigInteger.ONE).divide(BigInteger.valueOf(2));
    Group<BigInteger> group = new Zpstar(p);
    Arithmetic<BigInteger> arithmetic = new Fp(q);
    int t = 9;
    int n = 20;

    Testable[] testables;

    @Before
    public void settings(){
        testables = new Testable[tests];
        for (int i = 0; i < tests; i++){
            testables[i] = new Testable(new Random());
        }
    }

    public void oneTest(int test) throws Exception {
        Testable testable = testables[test];
        for (int i = 0; i < testable.threads.length ; i++){
            testable.threads[i].start();
        }
        for (int i = 0; i < testable.threads.length ; i++){
            testable.threads[i].join();
        }

        // got the right public value
        BigInteger publicValue = group.multiply(testable.g,testable.secret);
        for (int i: testable.valids){
            assert (testable.dkgs[i - 1].getPublicValue().equals(publicValue));
        }

        // assert valid verification values
        BigInteger expected,verification;
        for (int i: testable.valids){
            expected = group.multiply(testable.g, testable.dkgs[i - 1].getShare().y);
            verification = VerifiableSecretSharing.computeVerificationValue(i, testable.dkgs[i - 1].getCommitments(), group);
            assert (expected.equals(verification));
        }


        // restore the secret from shares
        ArrayList<Polynomial.Point> sharesList = new ArrayList<Polynomial.Point>();

        for (int i: testable.valids){
            sharesList.add(testable.dkgs[i - 1].getShare());
        }
        Polynomial.Point[] shares = new Polynomial.Point[sharesList.size()];
        for (int i = 0; i < shares.length; i ++){
            shares[i] = sharesList.get(i);
        }

        BigInteger calculatedSecret = SecretSharing.recoverSecret(shares,arithmetic);
        assert (calculatedSecret.equals(testable.secret));
    }

    @Test
    public void test() throws Exception {
        for (int i = 0; i < tests; i++){
            oneTest(i);
        }
    }

    class Testable{
        Set<Integer> valids;
        Set<Integer> QUAL;
        Set<Integer> aborted;
        Set<Integer> malicious;
        DistributedKeyGenerationUser[] dkgs;
        Thread[] threads;
        BigInteger g;
        BigInteger secret;

        public Testable(Random random) {
            this.dkgs = new DistributedKeyGenerationUserImpl[n];
            this.valids = new HashSet<Integer>();
            this.QUAL = new HashSet<Integer>();
            this.aborted = new HashSet<Integer>();
            this.malicious = new HashSet<Integer>();
            this.threads = new Thread[n];
            this.g = sampleGenerator(random);
            ArrayList<Integer> ids = new ArrayList<Integer>();
            for (int id = 1; id<= n ; id++){
                ids.add(id);
            }
            Network network = new Network(n);
            int id;
            BigInteger s;
            DistributedKeyGeneration dkg;
            this.secret = BigInteger.ZERO;
            while (!ids.isEmpty()) {
                id = ids.remove(random.nextInt(ids.size()));
                s = randomIntModQ(random);
                dkg = new DistributedKeyGeneration(t, n, s, random, q, g, group, id);
                dkgs[id - 1] = randomDKGUser(id,network,dkg,random);
                threads[id - 1] = new Thread(dkgs[id - 1]);
                if(QUAL.contains(id)){
                    this.secret =  this.secret.add(s).mod(q);
                }
            }

        }

        public DistributedKeyGenerationUser randomDKGUser(int id,Network network, DistributedKeyGeneration dkg,Random random){
            if (QUAL.size() <= t) {
                valids.add(id);
                QUAL.add(id);
                return new DistributedKeyGenerationUserImpl(dkg,network);
            }else{
                int type = random.nextInt(3);
                switch (type){
                    case 0:// regular
                        valids.add(id);
                        QUAL.add(id);
                        return new DistributedKeyGenerationUserImpl(dkg,network);
                    case 1:// abort
                        int abortStage = random.nextInt(2) + 1; // 1 or 2
                        aborted.add(id);
                        if (abortStage == 2){
                            QUAL.add(id);
                        }
                        return new DKGUserImplAbort(dkg,network,abortStage);
                    case 2:// malicious
                        malicious.add(id);
                        Set<Integer> falls = DKGMaliciousUserImpl.selectFallsRandomly(valids,random);
                        DistributedKeyGeneration maliciousDKG = DKGMaliciousUserImpl.generateMaliciousDKG(dkg,random);
                        return new DKGMaliciousUserImpl(dkg,maliciousDKG,network,falls);
                    default:
                        return null;
                }
            }
        }

        public BigInteger sampleGenerator(Random random){
            BigInteger ZERO = group.zero();
            BigInteger g;
            do {
                g = group.sample(random);
            } while (!g.equals(ZERO) && !group.multiply(g, q).equals(ZERO));
            return g;
        }

        public BigInteger randomIntModQ(Random random){
            return new BigInteger(q.bitLength(), random).mod(q);
        }

    }
}
